#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netpacket/packet.h>
#include <linux/types.h>
#include <linux/filter.h>

int main(int argc, char* argv[])
{
	int flags;
	int sockfd;
	int portnum;
	int debug = (2 <= argc && 0 == strcmp("-d", argv[1]));
	
	if (2 > argc || 0 >= (portnum = atoi(argv[argc - 1])))
	{
		fprintf(stderr, "Usage: %s <-d> udp_portnum\n", argv[0]);
		return -1;
	}
	
	if (0 <= (sockfd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_IP))) &&
	    0 <= (flags = fcntl(sockfd, F_GETFL)) &&
	    0 <= fcntl(sockfd, F_SETFL, flags | O_NONBLOCK))
	{
		struct sock_fprog filter;
		struct sock_filter BPF_code[]=
		{
			/* tcpdump -i lo udp and dst port 698 -dd */
			{ 0x28, 0, 0, 0x0000000c },
			{ 0x15, 0, 4, 0x000086dd },
			{ 0x30, 0, 0, 0x00000014 },
			{ 0x15, 0, 11, 0x00000011 },
			{ 0x28, 0, 0, 0x00000038 },
			{ 0x15, 8, 9, portnum },
			{ 0x15, 0, 8, 0x00000800 },
			{ 0x30, 0, 0, 0x00000017 },
			{ 0x15, 0, 6, 0x00000011 },
			{ 0x28, 0, 0, 0x00000014 },
			{ 0x45, 4, 0, 0x00001fff },
			{ 0xb1, 0, 0, 0x0000000e },
			{ 0x48, 0, 0, 0x00000010 },
			{ 0x15, 0, 1, portnum },
			{ 0x6, 0, 0, 0x00000060 },
			{ 0x6, 0, 0, 0x00000000 },
		};
		filter.len = sizeof(BPF_code) / sizeof(BPF_code[0]);
		filter.filter = BPF_code;
		if (0 > setsockopt(sockfd, SOL_SOCKET, SO_ATTACH_FILTER, &filter, sizeof(filter)))
		{
			perror("setsockopt");
			close(sockfd);
			return -1;
		}
	}
	else
	{
		perror("socket");
		return -1;
	}

	if (debug || 0 <= daemon(0, 0))
	{
		while(1)
		{
			struct
			{
				struct ethhdr eth;
				struct iphdr ip;
				struct udphdr udp;
			} __attribute__((packed)) buf;
			struct sockaddr_ll from;
			socklen_t fromlen = sizeof(from);
			
			ssize_t size = recvfrom(sockfd, &buf, sizeof(buf),
				MSG_TRUNC, (struct sockaddr *)&from, &fromlen);
		
			if (0 <= size)
			{
printf("size:%d\n", size);
				if (size >= sizeof(buf) &&
					ETH_P_IP == ntohs(buf.eth.h_proto) &&
					IPPROTO_UDP == buf.ip.protocol &&
					portnum == ntohs(buf.udp.source) &&
					portnum == ntohs(buf.udp.dest))
				{
					union
					{
						struct arpreq arp;
						struct sockaddr_in in_pa;
						struct sockaddr_in6 in_pa6;
					} req;
					
					memset(&req, 0, sizeof(req));
					
					req.in_pa.sin_family = AF_INET;
					memcpy(&req.in_pa.sin_addr, &buf.ip.saddr, sizeof(buf.ip.saddr));
					req.arp.arp_ha.sa_family = AF_LOCAL;
					memcpy(&req.arp.arp_ha.sa_data, &buf.eth.h_source, sizeof(buf.eth.h_source));
					req.arp.arp_flags = ATF_COM;
					if_indextoname(from.sll_ifindex, req.arp.arp_dev);
					
					if (debug)
					{
						int i;
						printf("Refresh on %s, %s=", req.arp.arp_dev,
							inet_ntoa(*((struct in_addr*)&buf.ip.saddr)));
						for(i = 0; i < sizeof(buf.eth.h_source); i++)
						{
							printf("%02x%s", ((unsigned char *)&buf.eth.h_source)[i], i < sizeof(buf.eth.h_source) -1 ? ":" : "\n");
						}
					}
					
					if (ioctl(sockfd, SIOCSARP, &req) < 0) {
						perror("SIOCSARP");
						close(sockfd);
						return -1;
					}
				}
			}
			else if (EAGAIN != errno)
			{
				perror("recv");
			}
		} /* while */
	}
	close(sockfd);
	return 0;
}
